﻿/* Gate Network Simulator
 * Copyright (c) 2008 Petr Koupy <petr.koupy@gmail.com>
 */

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;

namespace GateNetwork
{
    enum Kind
    {
        EOF,                // Peek() returns -1
        EOL,                // \n | \r | \n\r | \r\n
        Label,              // not contains white spaces, '.', ';', "->", not begins with "end"
        Gate,               // gate
        Composite,          // composite
        Network,            // network
        Inputs,             // inputs
        Outputs,            // outputs
        End,                // end
        Dot,                // .
        Arrow,              // ->
        Value               // 0 | 1 | ?
    }

    enum LogicValue
    {
        False,              // 0
        True,               // 1
        Undefined           // ?
    }

    struct Token
    {
        public readonly Kind kind;
        public readonly string label;
        public readonly LogicValue value;

        private Token(Kind kind, string label)
        {
            this.kind = kind;
            if (kind == Kind.Value)
            {
                this.label = label;
                switch (label)
                {
                    case "0":
                        this.value = LogicValue.False;
                        break;
                    case "1":
                        this.value = LogicValue.True;
                        break;
                    default:
                        this.value = LogicValue.Undefined;
                        break;
                }
            }
            else
            {
                this.label = label;
                this.value = LogicValue.Undefined;
            }
        }

        static public Token FromLabel(Kind kind, string label)
        {
            return new Token(kind, label);
        }
    }

    /* Provides conversion of text stream into token stream of 
     * implemented gramatics.
     */
    class TokenStream
    {
        private StreamReader reader;
        private bool dash; // -

        private Token currentToken;
        public Token CurrentToken
        {
            get
            {
                return this.currentToken;
            }
        }

        private int currentLine;
        public int CurrentLine
        {
            get
            {
                return this.currentLine;
            }
        }

        public TokenStream(StreamReader reader)
        {
            this.reader = reader;
            this.dash = false;
            this.currentLine = 1;
            this.currentToken = Token.FromLabel(Kind.EOF, "EOF"); // initialization
            this.Next(); // first token of file
        }

        /* Updates current token with new token provided by scanner on demand.
         * New token is also returned as return value.
         */
        public Token Next()
        {
            this.currentToken = Scan(this.reader);
            return this.currentToken;
        }

        /* Reads comment from input stream and throws it away. Comment is terminated
         * by newline.
         */
        private void ScanComment(StreamReader reader)
        {
            while ((char)reader.Peek() != '\r' && (char)reader.Peek() != '\n')
                reader.Read();
        }

        /* Reads string that is terminated by white space, '.', ';' or "->"
         * from input stream and returns it.
         */
        private string ScanWord(StreamReader reader)
        {
            StringBuilder builder = new StringBuilder();
            while (!Char.IsWhiteSpace((char)reader.Peek()) && (char)reader.Peek() != '.' && (char)reader.Peek() != ';' && (char)reader.Peek() != '>')
            {
                char c = (char)reader.Read();
                if (c == '-' && (char)reader.Peek() == '>')
                    this.dash = true;
                else
                    builder.Append(c);
            }
            return builder.ToString();
        }

        /* Scan function reads input stream until new token is recognized,
         * created and returned.
         */
        private Token Scan(StreamReader reader)
        {
            while (reader.Peek() != -1)
            {
                // End of line is detected and returned. Other white spaces are thrown away.
                if (Char.IsWhiteSpace((char)reader.Peek()))
                {
                    char white = (char)reader.Read();
                    if (white == '\r')
                    {
                        if ((char)reader.Peek() == '\n')
                        {
                            reader.Read();
                            this.currentLine++;
                        }
                        if (this.currentToken.kind != Kind.EOL)
                            return Token.FromLabel(Kind.EOL, "\r\n");
                    }
                    else if (white == '\n')
                    {
                        if ((char)reader.Peek() == '\r')
                        {
                            reader.Read();
                            this.currentLine++;
                        }
                        if (this.currentToken.kind != Kind.EOL)
                            return Token.FromLabel(Kind.EOL, "\r\n");
                    }
                }
                // Semicolon determines start of comment. Comment is thrown away by auxiliary function.
                else if ((char)reader.Peek() == ';')
                {
                    ScanComment(reader);
                }
                // Dot
                else if ((char)reader.Peek() == '.')
                {
                    reader.Read();
                    return Token.FromLabel(Kind.Dot, ".");
                }
                // Arrow
                else if (this.dash == true && (char)reader.Peek() == '>')
                {
                    reader.Read();
                    this.dash = false;
                    return Token.FromLabel(Kind.Arrow, "->");
                }
                // Other cases are detected directly.
                else
                {
                    if (!Char.IsWhiteSpace((char)reader.Peek()))
                    {
                        string word = ScanWord(reader);
                        switch (word)
                        {
                            case "gate":
                                return Token.FromLabel(Kind.Gate, word);
                            case "composite":
                                return Token.FromLabel(Kind.Composite, word);
                            case "network":
                                return Token.FromLabel(Kind.Network, word);
                            case "inputs":
                                return Token.FromLabel(Kind.Inputs, word);
                            case "outputs":
                                return Token.FromLabel(Kind.Outputs, word);
                            case "end":
                                return Token.FromLabel(Kind.End, word);
                            case "1":
                            case "0":
                            case "?":
                                return Token.FromLabel(Kind.Value, word);
                            default: // Label cannot start with "end".
                                if (word.Length >= 3 && word[0] == 'e' && word[1] == 'n' && word[2] == 'd')
                                    throw new Exception(String.Format("Line {0}: Syntax error.", this.currentLine));
                                else
                                    return Token.FromLabel(Kind.Label, word);
                        }
                    }
                }
            }
            return Token.FromLabel(Kind.EOF, "EOF");
        }
    }

    /* Stores information about one particular connection between two gates
     * or composites from the view of the owner of this joint. Provides logic
     * state of this connection.
     */
    [Serializable]
    class Joint
    {
        public Joint connector; 
        public Composite owner;
        public LogicValue value;

        public Joint()
        {
            this.connector = null;
            this.owner = null;
            this.value = LogicValue.Undefined;
        }
        public Joint(LogicValue value)
        {
            this.connector = null;
            this.owner = null;
            this.value = value;
        }

        public Joint(Composite owner)
        {
            this.connector = null;
            this.owner = owner;
            this.value = LogicValue.Undefined;
        }

        public Joint(Composite owner, Joint connector)
        {
            this.connector = connector;
            this.owner = owner;
            this.value = LogicValue.Undefined;
        }

        public Joint(Composite owner, LogicValue value)
        {
            this.connector = null;
            this.owner = owner;
            this.value = value;
        }
    }

    // Composite of at least one gate.
    [Serializable]
    class Composite : ICloneable
    {
        // Network is not stabilised while clock is true.
        public static bool clock = false;

        // Composite can be cloned in order to provide its instance.
        public object Clone()
        {
            BinaryFormatter formatter = new BinaryFormatter();
            MemoryStream stream = new MemoryStream();
            formatter.Serialize(stream, this);
            stream.Position = 0;
            object deepCopy = formatter.Deserialize(stream);
            stream.Close();
            return deepCopy;
        }

        protected string name;
        public string Name
        {
            get
            {
                return this.name;
            }
            set
            {
                this.name = value;
            }
        }

        public Dictionary<string, int> inputLabels; // Mapping of input names to indexes of input joint list.
        public Dictionary<string, int> outputLabels; // Mapping of output names to indexes of output joint list.
        public List<Joint> inputConnectors;
        public List<Joint> outputConnectors;
        public List<Composite> parts; // Pointers to internal gates and composites.

        private void Init()
        {
            this.inputLabels = new Dictionary<string, int>();
            this.outputLabels = new Dictionary<string, int>();
            this.inputConnectors = new List<Joint>();
            this.outputConnectors = new List<Joint>();
            this.parts = new List<Composite>();
        }

        public Composite()
        {
            this.name = "";
            Init();
        }

        public Composite(string name)
        {
            this.name = name;
            Init();
        }

        // Recursively fetches all input values.
        public virtual void Fetch()
        {
            foreach (Joint input in inputConnectors)
            {
                if (input.connector != null && input.connector.value != input.value)
                    input.value = input.connector.value;
            }
            foreach (Composite part in parts)
                part.Fetch();
        }

        // Recursively computes output values.
        public virtual void Clock()
        {
            foreach (Composite part in parts)
                part.Clock();
            foreach (Joint output in outputConnectors)
            {
                if (output.connector != null)
                    output.value = output.connector.value;
            }
        }
    }

    // Compares two joint lists.
    [Serializable]
    class JointListComparer : IEqualityComparer<List<Joint>>
    {
        // Determines whether corresponding logic values are equal in given joint lists.
        public bool Equals(List<Joint> first, List<Joint> second)
        {
            if (first.Count == second.Count)
            {
                for (int i = 0; i < first.Count; i++)
                    if (first[i].value != second[i].value)
                        return false;
                return true;
            }
            else
                return false;
        }

        public int GetHashCode(List<Joint> list)
        {
            return list.ToString().ToLower().GetHashCode();
        }
    }

    // Gate is a special kind of composite, which contains transitional function.
    [Serializable]
    class Gate : Composite
    {
        /* Transition table between given series of input logic values to 
         * series of output logic values. Note that input values must be 
         * first extracted from corresponding joints.
         */
        public Dictionary<List<Joint>, List<LogicValue>> function;

        // Gate is stabilised if no input was changed since last cycle.
        protected bool isStabilised;
        public bool IsStabilised
        {
            get
            {
                return this.isStabilised;
            }
        }
        
        public Gate() : base()
        {
            this.function = new Dictionary<List<Joint>, List<LogicValue>>(new JointListComparer());
            this.isStabilised = false;
        }

        public Gate(string name) : base(name)
        {
            this.function = new Dictionary<List<Joint>, List<LogicValue>>(new JointListComparer());
            this.isStabilised = false;
        }

        /* Determines outputs of gate, which has no inputs.
         * Note that this property ends recursion in network evaluation.
         */
        public void Set()
        {
            if (this.inputLabels.Count == 0) // No inputs
            {
                this.isStabilised = true;
                if (this.function.ContainsKey(new List<Joint>())) // Gate contains static transition.
                {
                    List<LogicValue> result = this.function[this.inputConnectors];
                    for (int i = 0; i < result.Count; i++)
                        this.outputConnectors[i].value = result[i];
                }
                else // Gate doesn't contain transition.
                {
                    foreach (Joint j in this.outputConnectors)
                        j.value = LogicValue.False;
                }
            }
        }

        // Fetches input values and decides whether the gate is stabilised or not.
        public override void Fetch()
        {
            this.isStabilised = true;
            foreach (Joint input in inputConnectors)
            {
                if (input.connector != null && input.connector.value != input.value)
                {
                    input.value = input.connector.value;
                    this.isStabilised = false;
                }
            }
        }

        /* Computes new output values if the gate is not stabilised.
         * This computation takes one clock signal.
         */
        public override void Clock()
        {
            if (!this.isStabilised)
            {
                Composite.clock = true; // Whole network is not stabilised.
                if (this.function.ContainsKey(this.inputConnectors)) // Explicit function rule.
                {
                    List<LogicValue> result = this.function[this.inputConnectors];
                    for (int i = 0; i < result.Count; i++)
                        this.outputConnectors[i].value = result[i];
                }
                else // Implicit function behaviour.
                {
                    LogicValue value = LogicValue.False;
                    // Any undefined input causes all undefined outputs.
                    for (int i = 0; i < this.inputConnectors.Count; i++)
                    {
                        if (this.inputConnectors[i].value == LogicValue.Undefined)
                        {
                            value = LogicValue.Undefined;
                            break;
                        }
                    }
                    foreach (Joint j in this.outputConnectors)
                        j.value = value;
                }
            }
        }
    }

    // Builds gate network from parsed token stream.
    class NetworkBuilder
    {
        private TokenStream tokenStream;
        private string[] errorMessage;

        public Composite network; // Gate network handle

        public static int maxCycles = 1000000; // Cycle in network would cause infinite loop.

        // The number of clock signals that was needed for network stabilisation.
        private int cycles;
        public int Cycles
        {
            get
            {
                return this.cycles;
            }
        }

        // Following two gates are connected to network inputs according to user request.
        public Gate TrueGate; // Implicit gate with zero inputs and one true output.
        public Gate FalseGate; // Implicit gate with zero inputs and one false output.

        public NetworkBuilder(TokenStream tokenStream)
        {
            this.tokenStream = tokenStream;
            this.network = new Composite();
            this.cycles = 0;
            this.errorMessage = new string[4];
            this.errorMessage[0] = "Duplicate.";
            this.errorMessage[1] = "Missing keyword.";
            this.errorMessage[2] = "Binding rule broken.";
            this.errorMessage[3] = "Syntax error.";

            // Initialize implicit true gate.
            this.TrueGate = new Gate();
            List<LogicValue> constantTrue = new List<LogicValue>();
            constantTrue.Add(LogicValue.True); // Static rule of transitional function.
            this.TrueGate.function.Add(new List<Joint>(), constantTrue); // Empty input corresponds to constant output.
            this.TrueGate.outputConnectors.Add(new Joint(this.TrueGate)); // Single output added.
            this.TrueGate.Set(); // Sets logic value stored in output connector according to internal transitional function.

            // Initialize implicit false gate.
            this.FalseGate = new Gate();
            List<LogicValue> constantFalse = new List<LogicValue>();
            constantFalse.Add(LogicValue.False);
            this.FalseGate.function.Add(new List<Joint>(), constantFalse);
            this.FalseGate.outputConnectors.Add(new Joint(this.FalseGate));
            this.FalseGate.Set();
        }

        // Order of enumeration should correspond to string error array in constructor.
        private enum Error
        {
            Duplicate,
            Keyword,
            Binding,
            Syntax
        }

        /* Checks whether current token is the same as expected given token.
         * If tokens are not the same, exception with given error is thrown.
         */
        private void Check(Kind kind, Error error)
        {
            if (this.tokenStream.CurrentToken.kind != kind)
            {
                string message = "Line " + this.tokenStream.CurrentLine + ": " + this.errorMessage[(int)error];
                throw new Exception(message);
            }
        }

        /* Checks whether given label is already contained in given dictionary.
         * If duplicate is found, exception with given error is thrown.
         */
        private void Check(string label, Dictionary<string, int> dictionary, Error error)
        {
            if (dictionary.ContainsKey(label))
            {
                string message = "Line " + this.tokenStream.CurrentLine + ": " + this.errorMessage[(int)error];
                throw new Exception(message);
            }
        }

        /* Checks whether given label is already contained in given dictionary.
        * If duplicate is found, exception with given error is thrown.
        */
        private void Check(string label, Dictionary<string, Composite> dictionary, Error error)
        {
            if (dictionary.ContainsKey(label))
            {
                string message = "Line " + this.tokenStream.CurrentLine + ": " + this.errorMessage[(int)error];
                throw new Exception(message);
            }
        }

        // Reads all input labels and saves them into given network element.
        private void Inputs(Composite element, TokenStream ts, bool isNetwork)
        {
            Check(Kind.Inputs, Error.Keyword);
            ts.Next();
            while (ts.CurrentToken.kind != Kind.EOL && ts.CurrentToken.kind != Kind.EOF)
            {
                Check(Kind.Label, Error.Syntax);
                Check(ts.CurrentToken.label, element.inputLabels, Error.Duplicate);
                // Build new input of given element.
                element.inputLabels.Add(ts.CurrentToken.label, element.inputConnectors.Count); // Name of input
                element.inputConnectors.Add(new Joint(element));
                ts.Next();
            }
            if (isNetwork && element.inputLabels.Count == 0) // Network must have at least one input.
                throw new Exception(String.Format("Line {0}: " + errorMessage[(int)Error.Syntax], ts.CurrentLine));
            Separator(ts);
        }

        // Reads all output labels and saves them into given network element.
        private void Outputs(Composite element, TokenStream ts)
        {
            Check(Kind.Outputs, Error.Keyword);
            ts.Next();
            while (ts.CurrentToken.kind != Kind.EOL && ts.CurrentToken.kind != Kind.EOF)
            {
                Check(Kind.Label, Error.Syntax);
                Check(ts.CurrentToken.label, element.outputLabels, Error.Duplicate);
                // Build new output of given element.
                element.outputLabels.Add(ts.CurrentToken.label, element.outputConnectors.Count); // Name of output
                element.outputConnectors.Add(new Joint(element));
                ts.Next();
            }
            if (element.outputLabels.Count == 0) // Any element must have at least one output.
                throw new Exception(String.Format("Line {0}: " + errorMessage[(int)Error.Syntax], ts.CurrentLine));
            Separator(ts);
        }

        // Reads explicit part of transitional function and saves it into given gate.
        private void Function(Gate gate, TokenStream ts)
        {
            while (ts.CurrentToken.kind == Kind.Value)
            {
                List<Joint> input = new List<Joint>();
                for (int i = 0; i < gate.inputLabels.Count; i++)
                {
                    Check(Kind.Value, Error.Syntax);
                    input.Add(new Joint(ts.CurrentToken.value));
                    ts.Next();
                }
                List<LogicValue> output = new List<LogicValue>();
                for (int i = 0; i < gate.outputLabels.Count; i++)
                {
                    Check(Kind.Value, Error.Syntax);
                    output.Add(ts.CurrentToken.value);
                    ts.Next();
                }
                if (gate.function.ContainsKey(input))
                    throw new Exception(String.Format("Line {0}: " + errorMessage[(int)Error.Duplicate], ts.CurrentLine));
                gate.function.Add(input, output);
                Separator(ts);
            }
        }

        /* Reads all labels and types of required composited and builds its instances.
         * Available types are read from given dictionary "types" and instances are saved
         * into given dictionary "instances".
         */
        private void Parts(Dictionary<string, Composite> types, Dictionary<string, Composite> instances, TokenStream ts)
        {
            while (ts.CurrentToken.kind == Kind.Gate)
            {
                ts.Next();
                Check(Kind.Label, Error.Syntax);
                string instanceName = ts.CurrentToken.label;
                Check(instanceName, instances, Error.Syntax); // Instance must have unique name.
                ts.Next();
                Check(Kind.Label, Error.Syntax);
                string typeName = ts.CurrentToken.label;
                if (!types.ContainsKey(typeName)) // No such type
                    throw new Exception(String.Format("Line {0}: " + errorMessage[(int)Error.Syntax], ts.CurrentLine));
                Composite instance = (Composite)types[typeName].Clone();
                instance.Name = instanceName;
                instances.Add(instanceName, instance);
                ts.Next();
                Separator(ts);
            }
        }

        /* Reads all bindings of given element and builds its internal structure from 
         * previously prepared instances. If given element is network itself, all connected
         * inputs are registered in given hashset in order to detect any unconnected input later.
         */
        private void Bindigs(Composite element, Dictionary<string, Composite> instances, TokenStream ts, bool isNetwork, HashSet<string> usedInputs)
        {
            while (ts.CurrentToken.kind == Kind.Label)
            {
                HashSet<string> patterns = new HashSet<string>(); // Serves for detection of duplicated bindings.
                string fromInstanceName = "";
                string fromConnectorName = "";
                string toInstanceName = "";
                string toConnectorName = "";
                string temp = "";
                temp = ts.CurrentToken.label;

                // Read left end of binding.
                ts.Next();
                if (ts.CurrentToken.kind != Kind.Dot)
                {
                    // End of binding is joint of given element.
                    Check(Kind.Arrow, Error.Syntax);
                    fromConnectorName = temp;
                }
                else
                {
                    // End of binding is joint of some instance.
                    fromInstanceName = temp;
                    ts.Next();
                    Check(Kind.Label, Error.Syntax);
                    fromConnectorName = ts.CurrentToken.label;
                    ts.Next();
                    Check(Kind.Arrow, Error.Syntax);
                }

                // Right end of binding can be '1' or '0', that are reserved labels for TrueGate and FalseGate.
                ts.Next();
                if (ts.CurrentToken.kind == Kind.Value && (ts.CurrentToken.value == LogicValue.True || ts.CurrentToken.value == LogicValue.False))
                    ;
                else
                    Check(Kind.Label, Error.Syntax);
                temp = ts.CurrentToken.label;

                // Read right end of binding.
                ts.Next();
                if (ts.CurrentToken.kind != Kind.Dot)
                {
                    Check(Kind.EOL, Error.Syntax);
                    toConnectorName = temp;
                }
                else
                {
                    toInstanceName = temp;
                    ts.Next();
                    Check(Kind.Label, Error.Syntax);
                    toConnectorName = ts.CurrentToken.label;
                    ts.Next();
                    Check(Kind.EOL, Error.Syntax);
                }
                Separator(ts);

                // Check whether binding is not duplicate of any previous.
                temp = fromInstanceName + fromConnectorName + toInstanceName + toConnectorName; // Unique ID of binding.
                if (patterns.Contains(temp))
                    throw new Exception(String.Format("Line {0}: " + errorMessage[(int)Error.Duplicate], ts.CurrentLine - 1));
                else
                    patterns.Add(temp);

                // Check whether binding is valid (input of any composite cannot be directly connected to its output).
                if (fromConnectorName != "" && toConnectorName != "" && !(fromInstanceName == "" && toInstanceName == ""))
                    ;
                else
                    throw new Exception(String.Format("Line {0}: " + errorMessage[(int)Error.Binding], ts.CurrentLine - 1));

                // Check whether binding is possible to realize (inputs, outputs and instances exists).
                Joint from = null;
                Joint to = null;
                if (fromInstanceName == "")
                {
                    if (!element.outputLabels.ContainsKey(fromConnectorName))
                        throw new Exception(String.Format("Line {0}: " + errorMessage[(int)Error.Binding], ts.CurrentLine - 1));
                    from = element.outputConnectors[element.outputLabels[fromConnectorName]];
                }
                else
                {
                    if (!instances.ContainsKey(fromInstanceName))
                        throw new Exception(String.Format("Line {0}: " + errorMessage[(int)Error.Binding], ts.CurrentLine - 1));
                    if (!instances[fromInstanceName].inputLabels.ContainsKey(fromConnectorName))
                        throw new Exception(String.Format("Line {0}: " + errorMessage[(int)Error.Binding], ts.CurrentLine - 1));
                    from = instances[fromInstanceName].inputConnectors[instances[fromInstanceName].inputLabels[fromConnectorName]];
                }
                if (toInstanceName == "" && toConnectorName == "1") // TrueGate
                {
                    to = this.TrueGate.outputConnectors[0];
                }
                else if (toInstanceName == "" && toConnectorName == "0") // FalseGate
                {
                    to = this.FalseGate.outputConnectors[0];
                }
                else if (toInstanceName == "")
                {
                    if (!element.inputLabels.ContainsKey(toConnectorName))
                        throw new Exception(String.Format("Line {0}: " + errorMessage[(int)Error.Binding], ts.CurrentLine - 1));
                    to = element.inputConnectors[element.inputLabels[toConnectorName]];
                    if (isNetwork)
                        usedInputs.Add(toConnectorName); // Detect connected network input.
                }
                else
                {
                    if (!instances.ContainsKey(toInstanceName))
                        throw new Exception(String.Format("Line {0}: " + errorMessage[(int)Error.Binding], ts.CurrentLine - 1));
                    if (!instances[toInstanceName].outputLabels.ContainsKey(toConnectorName))
                        throw new Exception(String.Format("Line {0}: " + errorMessage[(int)Error.Binding], ts.CurrentLine - 1));
                    to = instances[toInstanceName].outputConnectors[instances[toInstanceName].outputLabels[toConnectorName]];
                }
                
                /* Connect two joints. Note that whole internal structure of network is connected
                 * in the direction from outputs to inputs.
                 */
                if (from.connector == null)
                    from.connector = to;
                else
                    throw new Exception(String.Format("Line {0}: " + errorMessage[(int)Error.Binding], ts.CurrentLine - 1));
            }

            // Instances are added to internal reference list of network element.
            foreach (KeyValuePair<string, Composite> part in instances)
            {
                if (!element.parts.Contains(part.Value))
                    element.parts.Add(part.Value);
            }
        }

        // Checks whether the given element is connected properly.
        private void End(Composite element, TokenStream ts, bool isNetwork, HashSet<string> usedInputs)
        {
            Check(Kind.End, Error.Keyword);
            // Each output must be connected somewhere in element's internal structure.
            foreach (Joint j in element.outputConnectors)
            {
                if (j.connector == null)
                    throw new Exception(String.Format("Line {0}: " + errorMessage[(int)Error.Binding], ts.CurrentLine));
            }
            if (isNetwork)
            {
                // All inputs of network must be accessible somewhere from internal structure of network.
                foreach (string s in element.inputLabels.Keys)
                {
                    if (!usedInputs.Contains(s))
                        throw new Exception(String.Format("Line {0}: " + errorMessage[(int)Error.Binding], ts.CurrentLine));
                }
            }
            ts.Next();
        }

        // Reads all end of lines until any other token appears.
        private void Separator(TokenStream ts)
        {
            Check(Kind.EOL, Error.Syntax);
            while (ts.CurrentToken.kind == Kind.EOL)
                ts.Next();
        }

        // Recursively evaluates all outputs of network.
        public void Evaluate()
        {
            this.cycles = 0;
            do
            {
                Composite.clock = false;
                this.network.Fetch();
                this.network.Clock();
                if (Composite.clock)
                    this.cycles++;
            }
            while (Composite.clock && this.cycles < NetworkBuilder.maxCycles);
        }

        // Builds network.
        public Composite Build()
        {
            return build(this.tokenStream);
        }

        // Parses given stream of tokens and builds corresponding network.
        private Composite build(TokenStream ts)
        {
            // Dictionary of composite types. Any type must be instantiated before its connection to network.
            Dictionary<string, Composite> composites = new Dictionary<string, Composite>();

            while (ts.CurrentToken.kind != Kind.EOF)
            {
                switch (ts.CurrentToken.kind)
                {
                    case Kind.Gate:
                        ts.Next();
                        Check(Kind.Label, Error.Syntax);
                        Gate gate = new Gate(ts.CurrentToken.label);
                        ts.Next();
                        Separator(ts);
                        Inputs(gate, ts, false);
                        Outputs(gate, ts);
                        Function(gate, ts);
                        Check(Kind.End, Error.Keyword);
                        ts.Next();
                        Separator(ts);
                        gate.Set();
                        Check(gate.Name, composites, Error.Duplicate);
                        composites.Add(gate.Name, gate);
                        break;
                    case Kind.Composite:
                        Dictionary<string, Composite> instances1 = new Dictionary<string, Composite>();
                        ts.Next();
                        Check(Kind.Label, Error.Syntax);
                        Composite composite = new Composite(ts.CurrentToken.label);
                        ts.Next();
                        Separator(ts);
                        Inputs(composite, ts, false);
                        Outputs(composite, ts);
                        Check(Kind.Gate, Error.Keyword);
                        Parts(composites, instances1, ts);
                        Bindigs(composite, instances1, ts, false, null);
                        End(composite, ts, false, null);
                        Separator(ts);
                        Check(composite.Name, composites, Error.Duplicate);
                        composites.Add(composite.Name, composite);
                        break;
                    case Kind.Network:
                        if (composites.Count == 0) // There is no building block for network.
                            throw new Exception(String.Format("Line {0}: " + errorMessage[(int)Error.Syntax], ts.CurrentLine));
                        Dictionary<string, Composite> instances2 = new Dictionary<string, Composite>();
                        HashSet<string> usedInputs = new HashSet<string>();
                        ts.Next();
                        Separator(ts);
                        Inputs(this.network, ts, true);
                        Outputs(this.network, ts);
                        Check(Kind.Gate, Error.Keyword);
                        Parts(composites, instances2, ts);
                        Bindigs(this.network, instances2, ts, true, usedInputs);
                        End(this.network, ts, true, usedInputs);
                        ts.Next();
                        // Network definition must be last part of file.
                        if (ts.CurrentToken.kind == Kind.EOL)
                            Separator(ts);
                        Check(Kind.EOF, Error.Syntax);
                        this.network.Name = "OK"; // Determines whether there was network definition in a file.
                        break;
                    default:
                        ts.Next();
                        break;
                }
            }
            if (this.network.Name == "") // File not contains network definition.
                throw new Exception(String.Format("Line {0}: " + errorMessage[(int)Error.Syntax], ts.CurrentLine));
            return this.network;
        }
    }

    class Program
    {
        static void Main(string[] args)
        {
            StreamReader reader = null;
            try
            {
                // Only argument is a filename of source code file.
                if (args.Length == 1)
                    reader = new StreamReader(args[0]);
                else
                    throw new Exception("Invalid arguments.");

                NetworkBuilder networkBuilder = new NetworkBuilder(new TokenStream(reader));
                networkBuilder.Build();

                string inputLine = Console.ReadLine();
                while (inputLine != "end")
                {
                    string[] inputArgs = inputLine.Split(' ');
                    bool error = false;
                    if (inputArgs.Length == networkBuilder.network.inputConnectors.Count)
                    {
                        for (int i = 0; i < inputArgs.Length; i++)
                        {
                            if (inputArgs[i] != "1" && inputArgs[i] != "0")
                                error = true;
                        }
                    }
                    else
                        error = true;
                    if (!error)
                    {
                        // Connects TrueGate and FalseGate to corresponding inputs of network according to user input.
                        for (int i = 0; i < inputArgs.Length; i++)
                        {
                            if (inputArgs[i] == "1")
                                networkBuilder.network.inputConnectors[i].connector = networkBuilder.TrueGate.outputConnectors[0];
                            else
                                networkBuilder.network.inputConnectors[i].connector = networkBuilder.FalseGate.outputConnectors[0];
                        }

                        networkBuilder.Evaluate();

                        Console.Write(networkBuilder.Cycles);
                        for (int i = 0; i < networkBuilder.network.outputConnectors.Count; i++)
                        {
                            string output = "";
                            switch (networkBuilder.network.outputConnectors[i].value)
                            {
                                case LogicValue.False:
                                    output = " 0";
                                    break;
                                case LogicValue.True:
                                    output = " 1";
                                    break;
                                case LogicValue.Undefined:
                                    output = " ?";
                                    break;
                            }
                            Console.Write(output);
                        }
                        Console.WriteLine();
                    }
                    else
                        Console.WriteLine("Syntax error.");
                    inputLine = Console.ReadLine();
                }
            }
            catch (Exception exception)
            {
                Console.WriteLine(exception.Message);
                //Console.ReadLine();
            }
        }
    }
}
